/*
 *  sata_via.c - VIA Serial ATA controllers
 *
 *  Maintained by:  Jeff Garzik <jgarzik@pobox.com>
 * 		   Please ALWAYS copy linux-ide@vger.kernel.org
 *		   on emails.
 *
 *  Copyright 2003-2004 Red Hat, Inc.  All rights reserved.
 *  Copyright 2003-2004 Jeff Garzik
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  libata documentation is available via 'make {ps|pdf}docs',
 *  as Documentation/DocBook/libata.*
 *
 *  Hardware documentation available under NDA.
 *
 *
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <scsi/scsi_host.h>
#include <linux/libata.h>

#define DRV_NAME	"sata_via"
#define DRV_VERSION	"2.3"

/*
 * vt8251 is different from other sata controllers of VIA.
 * It has two channels, each channel has both Master and
 * Slave slot
 */
enum board_ids_enum {
	vt6420,
	vt6421,
	vt8251,
};

enum {
	SATA_CHAN_ENAB		= 0x40, /* SATA channel enable */
	SATA_INT_GATE		= 0x41, /* SATA interrupt gating */
	SATA_NATIVE_MODE	= 0x42, /* Native mode enable */
	PATA_UDMA_TIMING	= 0xB3, /* PATA timing for DMA/ cable detect */
	PATA_PIO_TIMING		= 0xAB, /* PATA timing register */

	PORT0			= (1 << 1),
	PORT1			= (1 << 0),
	ALL_PORTS		= PORT0 | PORT1,

	NATIVE_MODE_ALL		= (1 << 7) | (1 << 6) | (1 << 5) | (1 << 4),

	SATA_EXT_PHY		= (1 << 6), /* 0==use PATA, 1==ext phy */
};

static int svia_init_one(struct pci_dev *pdev, const struct pci_device_id *ent);
static int svia_scr_read(struct ata_port *ap, unsigned int sc_reg, u32 *val);
static int svia_scr_write(struct ata_port *ap, unsigned int sc_reg, u32 val);
static void svia_noop_freeze(struct ata_port *ap);

static void via_ata_tf_load(struct ata_port *ap, const struct ata_taskfile *tf);

static void via_std_set_pio_mode(struct ata_port *ap, struct ata_device *adev);
static void via_std_set_dma_mode(struct ata_port *ap, struct ata_device *adev);
static int via_std_cable_detect(struct ata_port *ap);

static int via_std_sata_prereset(struct ata_link *link,
			unsigned long deadline);
static int via_std_sata_comreset(struct ata_port *ap, int isMaster);

static const struct pci_device_id svia_pci_tbl[] = {
	{ PCI_VDEVICE(VIA, 0x3149), vt6420 }, /* Two sata channels(Master) */
	{ PCI_VDEVICE(VIA, 0x5337), vt6420 },
	{ PCI_VDEVICE(VIA, 0x5372), vt6420 },
	{ PCI_VDEVICE(VIA, 0x0591), vt6420 }, /* Two sata channels(Master) */
	{ PCI_VDEVICE(VIA, 0x7372), vt6420 },
	{ PCI_VDEVICE(VIA, 0x3249), vt6421 }, /* 2 sata chnls, 1 pata chnl*/
	{ PCI_VDEVICE(VIA, 0x5287), vt8251 }, /* 2 sata chnls(Master/Slave) */

	{ }	/* terminate list */
};

static struct pci_driver svia_pci_driver = {
	.name			= DRV_NAME,
	.id_table		= svia_pci_tbl,
	.probe			= svia_init_one,
#ifdef CONFIG_PM
	.suspend		= ata_pci_device_suspend,
	.resume			= ata_pci_device_resume,
#endif
	.remove			= ata_pci_remove_one,
};

static struct scsi_host_template svia_sht = {
	ATA_BMDMA_SHT(DRV_NAME),
};

/*
 * The opertion redefined by VIA to implement sata functions.
 * Since some via sata controllers support Master/Slave mode,
 * but the standard hardreset in libata could not work well
 * for slave slot, so we implement the hardreset during
 * prereset, and then do softreset. It should be mentioned
 * that libata does not do softreset for sata controller if
 * hardreset is availuable, to ensure the slave slot could
 * be detected, we skip standard hardrest in libata. That is
 * why we have to redefine the sata operations of sata
 */
static struct ata_port_operations via_std_sata_ops = {
	.inherits		= &ata_base_port_ops,

	.qc_prep		= ata_sff_qc_prep,
	.qc_issue		= ata_sff_qc_issue,
	.qc_fill_rtf		= ata_sff_qc_fill_rtf,

	.thaw			= ata_sff_thaw,
	.softreset		= ata_sff_softreset,
	.postreset		= ata_sff_postreset,
	.error_handler		= ata_sff_error_handler,
	.post_internal_cmd	= ata_sff_post_internal_cmd,

	.sff_dev_select		= ata_sff_dev_select,
	.sff_check_status	= ata_sff_check_status,
	.sff_tf_read		= ata_sff_tf_read,
	.sff_exec_command	= ata_sff_exec_command,
	.sff_data_xfer		= ata_sff_data_xfer,
	.sff_irq_on			= ata_sff_irq_on,
	.sff_irq_clear		= ata_sff_irq_clear,

	.port_start		= ata_sff_port_start,

	.mode_filter		= ata_bmdma_mode_filter,

	.bmdma_setup		= ata_bmdma_setup,
	.bmdma_start		= ata_bmdma_start,
	.bmdma_stop			= ata_bmdma_stop,
	.bmdma_status		= ata_bmdma_status,

	.sff_tf_load		= via_ata_tf_load,

	.prereset		= via_std_sata_prereset,
	.freeze			= svia_noop_freeze,

	.scr_read		= svia_scr_read,
	.scr_write		= svia_scr_write,
};

static struct ata_port_operations via_std_pata_ops = {
	.inherits		= &ata_bmdma_port_ops,

	.cable_detect	= via_std_cable_detect,
	.set_piomode	= via_std_set_pio_mode,
	.set_dmamode	= via_std_set_dma_mode,

	.sff_tf_load	= via_ata_tf_load,
};


static struct ata_port_info vt6420_port_info = {
	.flags		= ATA_FLAG_SATA | ATA_FLAG_NO_LEGACY,
	.pio_mask	= 0x1f,
	.mwdma_mask	= 0x07,
	.udma_mask	= ATA_UDMA6,
	.port_ops	= &via_std_sata_ops,
};

static struct ata_port_info vt6421_sport_info = {
	.flags		= ATA_FLAG_SATA | ATA_FLAG_NO_LEGACY,
	.pio_mask	= 0x1f,
	.mwdma_mask	= 0x07,
	.udma_mask	= ATA_UDMA6,
	.port_ops	= &via_std_sata_ops,
};

static struct ata_port_info vt6421_pport_info = {
	.flags		= ATA_FLAG_SLAVE_POSS | ATA_FLAG_NO_LEGACY,
	.pio_mask	= 0x1f,
	.mwdma_mask	= 0,
	.udma_mask	= ATA_UDMA6,
	.port_ops	= &via_std_pata_ops,
};

/**
 * VT8251 has two sata channels, which are both Master/Slave mode. This is different
 * from the implementation in libata, which supports only master mode
 */
static struct ata_port_info vt8251_port_info = {
	.flags		=  ATA_FLAG_SATA | ATA_FLAG_SLAVE_POSS |
					ATA_FLAG_NO_LEGACY,
	.pio_mask	= 0x1f,
	.mwdma_mask	= 0x07,
	.udma_mask	= ATA_UDMA6,
	.port_ops	= &via_std_sata_ops,
};

MODULE_AUTHOR("Jeff Garzik");
MODULE_DESCRIPTION("SCSI low-level driver for VIA SATA controllers");
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(pci, svia_pci_tbl);
MODULE_VERSION(DRV_VERSION);

/**
 * For controller 0x8251 and ox0581/0x5324, the sata controller registers do not
 * locate in the PCI BARs, instead, they are in the PCI configure space.
 */
static int svia_scr_read(struct ata_port *ap, unsigned int sc_reg, u32 *val)
{
	struct pci_dev *pdev = to_pci_dev(ap->host->dev);
	u8  config_space_value;
    u8  byte_temp;
	u32 scr_value;
    u32 dword_temp;
    s8  shift;
    u8  slot;

    unsigned int scr_reg = sc_reg;

    if (scr_reg > SCR_CONTROL) {
		VPRINTK("invalid scr[%d] address!\n", scr_reg);
		return -EINVAL;
    }


    if (!(pdev->device == 0x5287 || pdev->device == 0x0581 ||
		pdev->device == 0x5324))
		*val = ioread32(ap->ioaddr.scr_addr + (4 * scr_reg));
    else {
		config_space_value = 0;
		scr_value = 0;
		shift     = 0;

		/**
		 * Since these chips do not use iomap to get the address of scr,
		 * we use these register to store the slot index
		 */
		slot = (u32)ap->ioaddr.scr_addr & 0x03;

		if (scr_reg == SCR_STATUS) {
			pci_read_config_byte(pdev, (0xA0 + (slot & 0x03)),
				&config_space_value);

			/* Set the DET field, bit0 and 1 of the config byte */
			scr_value |= (config_space_value & 0x03);

			/* Set the SPD field, bit4 of the configure byte */
			if (config_space_value & (1<<4))
				scr_value |= (u8)(0x02 << 4);
			else
				scr_value |= (u8)(0x01 << 4);

			/* Set the IPM field, bit2 and 3 of the config byte */
			byte_temp = (config_space_value >> 2) & 0x03;
			shift = byte_temp - 1;
			if (shift < 0)
				shift = 0;

			dword_temp = (byte_temp + 0x01) << shift;
			scr_value  |= ((dword_temp & 0x0f) << 8);
		} else if (scr_reg == SCR_ERROR) {
			if (pdev->device == 0x5287)
				pci_read_config_dword(pdev,
				(0xB0 + (slot & 0x03) * 4), &scr_value);
			else
				pci_read_config_dword(pdev,
				(0xA8 + (slot & 0x03) * 4), &scr_value);
		} else if (scr_reg == SCR_CONTROL) {
			pci_read_config_byte(pdev,
				(0xA4 + (slot & 0x03)), &config_space_value);

			 /* Read the DET field, bit0 and bit1 */
			 byte_temp = config_space_value & 0x03;
			 scr_value |= (((byte_temp & 0x02) << 1) |
				 (byte_temp & 0x01));

			/* Read the SPD field, bit4 */
			scr_value &= ~(u32)0xF0;

			/* Read the IPM field, bit2 and bit3 */
			dword_temp = (config_space_value >> 2) & 0x03;
			scr_value  |= (dword_temp << 8);
		}

		*val = scr_value;
    }

    return 0;
}



static int svia_scr_write(struct ata_port *ap, unsigned int sc_reg, u32 val)
{
	struct pci_dev *pdev = to_pci_dev(ap->host->dev);
	u8  config_space_value;
    u8  byte_temp;
    u8  slot;

	unsigned int scr_reg = sc_reg;

    if (scr_reg > SCR_CONTROL) {
		VPRINTK("invalid scr[%d] address!\n", scr_reg);
		return -EINVAL;
    }

    if (scr_reg == SCR_STATUS) {
		VPRINTK("scr_status can not be write!\n");
		return -EINVAL;
    }

    if (!(pdev->device == 0x5287 || pdev->device == 0x0581 ||
		pdev->device == 0x5324))
		iowrite32(val, ap->ioaddr.scr_addr + (4 * scr_reg));
    else {
		byte_temp  = 0;
		config_space_value = 0;

		/* To get the slot index */
	    slot = (u32) ap->ioaddr.scr_addr & 0x03;

		if (scr_reg == SCR_ERROR) {
			if (pdev->device == 0x5287)
				pci_write_config_dword(pdev,
				(0xB0 + (slot & 0x03) * 4), val);
			else
				pci_write_config_dword(pdev,
				(0xA8 + (slot & 0x03) * 4), val);
		} else if (scr_reg == SCR_CONTROL) {
			/* Set the DET field */
			byte_temp = (((u8)(val & 0x04)) >> 1) |
				((u8)(val & 0x01));
			config_space_value |= (byte_temp & 0x0F);

			/* Set the IPM field */
			byte_temp = (val >> 8) & 0x03;
			config_space_value |= (byte_temp << 2);
			pci_write_config_byte(pdev, (0xA4 + (slot & 0x03)),
				config_space_value);
		}
    }

	return 0;
}

/**
 *	via_ata_sff_tf_load - send taskfile registers to host controller
 *	@ap: Port to which output is sent
 *	@tf: ATA taskfile register set
 *
 *	Outputs ATA taskfile to standard ATA host controller.
 *
 *	Note: This is to fix the internal bug of via chipsets,
 *  which will reset the device register after changing the IEN bit
 *  on ctl register
 */
static void via_ata_tf_load(struct ata_port *ap, const struct ata_taskfile *tf)
{
	struct ata_ioports *ioaddr = &ap->ioaddr;
	unsigned int is_addr = tf->flags & ATA_TFLAG_ISADDR;

	if (tf->ctl != ap->last_ctl) {
		iowrite8(tf->ctl, ioaddr->ctl_addr);
		iowrite8(tf->device, ioaddr->device_addr);
		ap->last_ctl = tf->ctl;
		ata_wait_idle(ap);
	}

	if (is_addr && (tf->flags & ATA_TFLAG_LBA48)) {
		iowrite8(tf->hob_feature, ioaddr->feature_addr);
		iowrite8(tf->hob_nsect, ioaddr->nsect_addr);
		iowrite8(tf->hob_lbal, ioaddr->lbal_addr);
		iowrite8(tf->hob_lbam, ioaddr->lbam_addr);
		iowrite8(tf->hob_lbah, ioaddr->lbah_addr);
		VPRINTK("hob: feat 0x%X nsect 0x%X, lba 0x%X 0x%X 0x%X\n",
			tf->hob_feature,
			tf->hob_nsect,
			tf->hob_lbal,
			tf->hob_lbam,
			tf->hob_lbah);
	}

	if (is_addr) {
		iowrite8(tf->feature, ioaddr->feature_addr);
		iowrite8(tf->nsect, ioaddr->nsect_addr);
		iowrite8(tf->lbal, ioaddr->lbal_addr);
		iowrite8(tf->lbam, ioaddr->lbam_addr);
		iowrite8(tf->lbah, ioaddr->lbah_addr);
		VPRINTK("feat 0x%X nsect 0x%X lba 0x%X 0x%X 0x%X\n",
			tf->feature,
			tf->nsect,
			tf->lbal,
			tf->lbam,
			tf->lbah);
	}

	if (tf->flags & ATA_TFLAG_DEVICE) {
		iowrite8(tf->device, ioaddr->device_addr);
		VPRINTK("device 0x%X\n", tf->device);
	}

	ata_wait_idle(ap);
}

static int via_std_cable_detect(struct ata_port *ap)
{
	struct pci_dev *pdev = to_pci_dev(ap->host->dev);
	u8 tmp;
	int cfg_addr;

	if (ap->port_no == 2) {
		if (pdev->device == 0x3249) {
			cfg_addr = 0xB3;
			pci_read_config_byte(pdev, cfg_addr, &tmp);
			if (tmp & 0x10)
				return ATA_CBL_PATA40;
			return ATA_CBL_PATA80;
		}
	} else if (ap->port_no == 1) {
		if ((pdev->device == 0x0581) || (pdev->device == 0x5324)) {
			cfg_addr = 0x50;
			pci_read_config_byte(pdev, cfg_addr, &tmp);
			if (tmp & 0x10) /* 40pin cable */
				return ATA_CBL_PATA40;
			else /* 80pin cable */
				return ATA_CBL_PATA80;
		}
	}

	return ATA_CBL_SATA;
}

static void svia_noop_freeze(struct ata_port *ap)
{
	/* Some VIA controllers choke if ATA_NIEN is manipulated in
	 * certain way.  Leave it alone and just clear pending IRQ.
	 */
	ap->ops->sff_check_status(ap);
	ata_sff_irq_clear(ap);
}

 /**
 *	via_std_sata_comreset - comreset for via sata controller
 *	@ap: target ATA port
 *	@isMaster: whether the input port is master slot
 *
 *	Some VIA controllers does not use PCI BAR to map the sata
 *  controller registers, besides, some via data controllers
 *  have both master and slave slots. As a result, standard
 *  operations in libata could not work well
 *
 *	LOCKING:
 *	Kernel thread context (may sleep)
 *
 *	RETURNS:
 *	0 on success, -errno otherwise.
 */
static int via_std_sata_comreset(struct ata_port *ap, int isMaster)
{
    struct pci_dev *pdev = to_pci_dev(ap->host->dev);
	unsigned long timeout = jiffies + (HZ * 5);
	u32 sstatus, scontrol, serror;
	int online = 0;

	int slot;
    int slot_start = ap->port_no;
	unsigned int dev = 0;

	unsigned int cbl_tmp = via_std_cable_detect(ap);
	unsigned long flag_tmp = ap->flags;

	if (ap->flags & ATA_FLAG_SLAVE_POSS)
		slot_start = slot_start << 1;

	ap->cbl = ATA_CBL_SATA;
	ap->flags |= ATA_FLAG_SATA;

	{
		slot = (isMaster ? 0 : 1) + slot_start;
		if (pdev->device == 0x5287 || pdev->device == 0x0581 ||
			pdev->device == 0x5324)
			ap->ioaddr.scr_addr = (void __iomem *)slot;

		dev =  (ap->flags & ATA_FLAG_SLAVE_POSS) ? (slot&0x01) : 0;
		ap->ops->sff_dev_select(ap, dev);

		DPRINTK("Comreset SATA Device %X PORT : %d, SLOT : %s\n", \
			    pdev->device, ap->port_no, \
				(dev == 0) ? "<Master>" : "<Slave>");

		ap->ops->scr_read(ap, SCR_CONTROL, &scontrol);
		scontrol = (scontrol & 0x0f0) | 0x301;
		ap->ops->scr_write(ap, SCR_CONTROL, scontrol);
		ap->ops->scr_read(ap, SCR_CONTROL, &scontrol);

		msleep(2);
		scontrol = (scontrol & 0x0f0) | 0x300;
		ap->ops->scr_write(ap, SCR_CONTROL, scontrol);
		ap->ops->scr_read(ap, SCR_CONTROL, &scontrol); /* flush */

		/* wait for phy to become ready, if necessary */
		do {
			msleep(200);
			ap->ops->scr_read(ap, SCR_STATUS, &sstatus);
			if ((sstatus & 0xf) != 1)
				break;
		} while (time_before(jiffies, timeout));

		/* open code sata_print_link_status() */
		ap->ops->scr_read(ap, SCR_STATUS, &sstatus);
		ap->ops->scr_read(ap, SCR_CONTROL, &scontrol);
		ap->ops->scr_read(ap, SCR_ERROR, &serror);
		ap->ops->scr_write(ap, SCR_ERROR, serror);

		ata_port_printk(ap, KERN_INFO,
				"SATA link %s (SStatus %X SControl %X)\n",
				online ? "up" : "down", sstatus, scontrol);

		/* SStatus is read one more time */
		ap->ops->scr_read(ap, SCR_STATUS, &sstatus);

		online = (sstatus & 0xf) == 0x3;
		if (!online) {
			/* tell EH to bail */
			DPRINTK("Comreset %X PORT : %d, SLOT : %s  failed\n", \
				   pdev->device, ap->port_no, \
					(dev == 0) ? "<Master>" : "<Slave>");

			ap->cbl = cbl_tmp;
			ap->flags = flag_tmp;
			return -ENODEV;
		}
		DPRINTK("Comreset %X PORT : %d, SLOT : %s succeeded.\n", \
			     pdev->device, ap->port_no, \
				(dev == 0) ? "<Master>" : "<Slave>");

	}

	ap->cbl = cbl_tmp;
	ap->flags = flag_tmp;


	return 0;
}


/**
 *	via_std_sata_prereset - prereset for via sata controllers
 *	@link: target ATA link
 *	@deadline: deadline jiffies for the operation
 *
 *	Standard error handling will retry several times if reset
 *  fails, this makes users cannot bear it.
 *
 *	LOCKING:
 *	Kernel thread context (may sleep)
 *
 *	RETURNS:
 *	0 on success, -errno otherwise.
 */
static int via_std_sata_prereset(struct ata_link *link, unsigned long deadline)
{
	struct ata_port *ap = link->ap;
    struct pci_dev *pdev = to_pci_dev(ap->host->dev);
	int devmask = 0;
	int slot_cnt = 0;
	int slot;
	int rc;
	u8 status;

	/* don't do any SCR stuff if we're not loading */
	if (!(ap->pflags & ATA_PFLAG_LOADING))
		goto skip_scr;

	slot_cnt = (ap->flags & ATA_FLAG_SLAVE_POSS) ? 2 : 1;
	for (slot = 0; slot < slot_cnt; ++slot) {
		rc = via_std_sata_comreset(ap, (slot%2 == 0) ? 1 : 0);
		if (!rc)
			devmask |= (1<<slot);

		rc = ata_sff_wait_ready(link, deadline);
		status = link->ap->ops->sff_check_status(link->ap);
		DPRINTK("Prereset %X for PORT%d status %x: \n", \
				pdev->device, ap->port_no, status);
		if (rc != 0) {
			DPRINTK("Prereset %X for PORT%d failed\n", \
						pdev->device, ap->port_no);
			devmask &= ~(1<<slot);
		}
	}

	/*
	 * To fix the online check in libata. The softreset in libata will check
	 * whether the Master slot will be online for an ata link, so we should
	 * reset the master port again if the slave slot does not have a device
	 * attached
	 */
	if ((slot_cnt == 2) && (devmask == 0x01))
		via_std_sata_comreset(ap, 1);
	else if (devmask == 0) {
			DPRINTK("Prereset%X for PORT%d failed: (No device)\n", \
					pdev->device, ap->port_no);
			return -ENODEV;
	}

skip_scr:
	/* wait for !BSY */
	rc = ata_sff_wait_ready(link, deadline);
	if (rc)
		DPRINTK("Prereset %X for PORT%d failed\n", \
				    pdev->device, ap->port_no);

	return rc;
}


static void via_std_set_pio_mode(struct ata_port *ap, struct ata_device *adev)
{
	struct pci_dev *pdev = to_pci_dev(ap->host->dev);

	u8 cfg_byte;
	int cfg_addr;
	u16 tmp16;

	tmp16 = pdev->device;

	if ((tmp16 == 0x3249) && (ap->port_no == 2))/* PATA channel in VT6421 */
		cfg_addr = 0xAB;
	else if ((tmp16 == 0x0581 || tmp16 == 0x5324) &&
		(ap->port_no == 1))/* PATA channel in VT8353 */
		cfg_addr = 0x49;
	else
		return;

	switch (adev->pio_mode & 0x07) {
	case 0:
		cfg_byte = 0xa8;
		break;
	case 1:
		cfg_byte = 0x65;
		break;
	case 2:
		cfg_byte = 0x65;
		break;
	case 3:
		cfg_byte = 0x31;
		break;
	case 4:
		cfg_byte = 0x20;
		break;
	default:
		cfg_byte = 0x20;
	}

	pci_write_config_byte(pdev, cfg_addr, cfg_byte);
}

static void via_std_set_dma_mode(struct ata_port *ap, struct ata_device *adev)
{
	struct pci_dev *pdev = to_pci_dev(ap->host->dev);
	u8 cfg_byte;
	int cfg_addr;
	u16 tmp16;

	tmp16 = pdev->device;

	if ((tmp16 == 0x3249) && (ap->port_no == 2))
		cfg_addr = 0xB3; /* PATA channel in VT6421 */
	else if ((tmp16 == 0x0581) && (ap->port_no == 1))
		cfg_addr = 0x50; /* PATA channel in VT8353 */
	else
		return;

	switch (adev->dma_mode & 0x07) {
	case 0:
		cfg_byte = 0xee;
		break;
	case 1:
		cfg_byte = 0xe8;
		break;
	case 2:
		cfg_byte = 0xe6;
		break;
	case 3:
		cfg_byte = 0xe4;
		break;
	case 4:
		cfg_byte = 0xe2;
		break;
	case 5:
		cfg_byte = 0xe1;
		break;
	case 6:
		cfg_byte = 0xe0;
		break;
	default:
		cfg_byte = 0xe0;
	}

	pci_write_config_byte(pdev, cfg_addr, cfg_byte);
}

static const unsigned int svia_bar_sizes[] = {
	8, 4, 8, 4, 16, 256
};

static const unsigned int vt6421_bar_sizes[] = {
	16, 16, 16, 16, 32, 128
};

static void __iomem *svia_scr_addr(void __iomem *addr, unsigned int port)
{
	return addr + (port * 128);
}

static void __iomem *vt6421_scr_addr(void __iomem *addr, unsigned int port)
{
	return addr + (port * 64);
}

static void vt6421_init_addrs(struct ata_port *ap)
{
	void __iomem * const *iomap = ap->host->iomap;
	void __iomem *reg_addr = iomap[ap->port_no];
	void __iomem *bmdma_addr = iomap[4] + (ap->port_no * 8);
	struct ata_ioports *ioaddr = &ap->ioaddr;

	ioaddr->cmd_addr = reg_addr;
	ioaddr->altstatus_addr =
	ioaddr->ctl_addr = (void __iomem *)
		((unsigned long)(reg_addr + 8) | ATA_PCI_CTL_OFS);
	ioaddr->bmdma_addr = bmdma_addr;
	ioaddr->scr_addr = vt6421_scr_addr(iomap[5], ap->port_no);

	ata_sff_std_ports(ioaddr);

	ata_port_pbar_desc(ap, ap->port_no, -1, "port");
	ata_port_pbar_desc(ap, 4, ap->port_no * 8, "bmdma");
}

static int vt6420_prepare_host(struct pci_dev *pdev, struct ata_host **r_host)
{
	const struct ata_port_info *ppi[] = { &vt6420_port_info, NULL };
	struct ata_host *host;
	int rc;

	rc = ata_pci_sff_prepare_host(pdev, ppi, &host);
	if (rc)
		return rc;
	*r_host = host;

	rc = pcim_iomap_regions(pdev, 1 << 5, DRV_NAME);
	if (rc) {
		dev_printk(KERN_ERR, &pdev->dev, "failed to iomap PCI BAR 5\n");
		return rc;
	}

	host->ports[0]->ioaddr.scr_addr = svia_scr_addr(host->iomap[5], 0);
	host->ports[1]->ioaddr.scr_addr = svia_scr_addr(host->iomap[5], 1);

	return 0;
}

static int vt6421_prepare_host(struct pci_dev *pdev, struct ata_host **r_host)
{
	const struct ata_port_info *ppi[] =
		{ &vt6421_sport_info, &vt6421_sport_info, &vt6421_pport_info };
	struct ata_host *host;
	int i, rc;

	*r_host = host = ata_host_alloc_pinfo(&pdev->dev, ppi, ARRAY_SIZE(ppi));
	if (!host) {
		dev_printk(KERN_ERR, &pdev->dev, "failed to allocate host\n");
		return -ENOMEM;
	}

	rc = pcim_iomap_regions(pdev, 0x3f, DRV_NAME);
	if (rc) {
		dev_printk(KERN_ERR, &pdev->dev, "failed to request/iomap "
			   "PCI BARs (errno=%d)\n", rc);
		return rc;
	}
	host->iomap = pcim_iomap_table(pdev);

	for (i = 0; i < host->n_ports; i++)
		vt6421_init_addrs(host->ports[i]);

	rc = pci_set_dma_mask(pdev, ATA_DMA_MASK);
	if (rc)
		return rc;
	rc = pci_set_consistent_dma_mask(pdev, ATA_DMA_MASK);
	if (rc)
		return rc;

	return 0;
}

static int vt8251_prepare_host(struct pci_dev *pdev, struct ata_host **r_host)
{
    const struct ata_port_info *ppi[] = {&vt8251_port_info, NULL};
	struct ata_host *host;
	int rc;

	rc = ata_pci_sff_prepare_host(pdev, ppi, &host);
	if (rc)
		return rc;
	*r_host = host;

	rc = pcim_iomap_regions(pdev, 1 << 5, DRV_NAME);
	if (rc) {
		dev_printk(KERN_ERR, &pdev->dev, "failed to iomap PCI BAR 5\n");
		return rc;
	}

	return 0;
}

static void svia_configure(struct pci_dev *pdev)
{
	u8 tmp8;

	pci_read_config_byte(pdev, PCI_INTERRUPT_LINE, &tmp8);
	dev_printk(KERN_INFO, &pdev->dev, "routed to hard irq line %d\n",
	       (int) (tmp8 & 0xf0) == 0xf0 ? 0 : tmp8 & 0x0f);

	/* make sure SATA channels are enabled */
	pci_read_config_byte(pdev, SATA_CHAN_ENAB, &tmp8);
	if ((tmp8 & ALL_PORTS) != ALL_PORTS) {
		dev_printk(KERN_DEBUG, &pdev->dev,
			   "enabling SATA channels (0x%x)\n",
			   (int) tmp8);
		tmp8 |= ALL_PORTS;
		pci_write_config_byte(pdev, SATA_CHAN_ENAB, tmp8);
	}

	/* make sure interrupts for each channel sent to us */
	pci_read_config_byte(pdev, SATA_INT_GATE, &tmp8);
	if ((tmp8 & ALL_PORTS) != ALL_PORTS) {
		dev_printk(KERN_DEBUG, &pdev->dev,
			   "enabling SATA channel interrupts (0x%x)\n",
			   (int) tmp8);
		tmp8 |= ALL_PORTS;
		pci_write_config_byte(pdev, SATA_INT_GATE, tmp8);
	}

	/* make sure native mode is enabled */
	pci_read_config_byte(pdev, SATA_NATIVE_MODE, &tmp8);
	if ((tmp8 & NATIVE_MODE_ALL) != NATIVE_MODE_ALL) {
		dev_printk(KERN_DEBUG, &pdev->dev,
			   "enabling SATA channel native mode (0x%x)\n",
			   (int) tmp8);
		tmp8 |= NATIVE_MODE_ALL;
		pci_write_config_byte(pdev, SATA_NATIVE_MODE, tmp8);
	}
}

static int svia_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	static int printed_version;
	unsigned int i;
	int rc;
	struct ata_host *host;
	int board_id = (int) ent->driver_data;
	const unsigned *bar_sizes;

	if (!printed_version++)
		dev_printk(KERN_DEBUG, &pdev->dev, "version " DRV_VERSION "\n");

	rc = pcim_enable_device(pdev);
	if (rc)
		return rc;

	if (board_id == vt6420)
		bar_sizes = &svia_bar_sizes[0];
	else if (board_id == vt6421)
		bar_sizes = &vt6421_bar_sizes[0];
	else
		bar_sizes = &svia_bar_sizes[0];


	for (i = 0; i < ARRAY_SIZE(svia_bar_sizes); i++)
		if ((pci_resource_start(pdev, i) == 0) ||
		    (pci_resource_len(pdev, i) < bar_sizes[i])) {
			dev_printk(KERN_ERR, &pdev->dev,
				"invalid PCI BAR %u (sz 0x%llx, val 0x%llx)\n",
				i,
				(unsigned long long)pci_resource_start(pdev, i),
				(unsigned long long)pci_resource_len(pdev, i));
			return -ENODEV;
		}

	if (board_id == vt6420)
		rc = vt6420_prepare_host(pdev, &host);
	else if (board_id == vt6421)
		rc = vt6421_prepare_host(pdev, &host);
	else
		rc = vt8251_prepare_host(pdev, &host);

	if (rc)
		return rc;

	svia_configure(pdev);

	pci_set_master(pdev);
	return ata_host_activate(host, pdev->irq, ata_sff_interrupt,
				 IRQF_SHARED, &svia_sht);
}

static int __init svia_init(void)
{
	return pci_register_driver(&svia_pci_driver);
}

static void __exit svia_exit(void)
{
	pci_unregister_driver(&svia_pci_driver);
}

module_init(svia_init);
module_exit(svia_exit);
