/*
 * PIKA FPGA based Watchdog Timer
 *
 * Copyright (c) 2008 PIKA Technologies
 *   Sean MacLennan <smaclennan@pikatech.com>
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/reboot.h>
#include <linux/uaccess.h>
#include <linux/io.h>
#include <linux/of_platform.h>
#include <linux/watchdog.h>


static void __iomem *pikawdt_fpga;


static inline void pikawdt_ping(void)
{
	unsigned reset = in_be32(pikawdt_fpga + 0x14);
	reset |= 0xf80; /* enable with max timeout - 15 seconds */
	out_be32(pikawdt_fpga + 0x14, reset);
}

static int pikawdt_open(struct inode *inode, struct file *file)
{
	printk(KERN_INFO "PIKA WDT started...\n");

	pikawdt_ping();

	return 0;
}

static int pikawdt_release(struct inode *inode, struct file *file)
{
	pikawdt_ping(); /* one last time */
	return 0;
}

static ssize_t pikawdt_write(struct file *file, const char __user *buf,
			     size_t count, loff_t *ppos)
{
	pikawdt_ping();
	return count;
}

/* We support the bare minimum to be conformant. */
static int pikawdt_ioctl(struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg)
{
	if (cmd == WDIOC_KEEPALIVE) {
		pikawdt_ping();
		return 0;
	} else
		return -EINVAL;
}

static const struct file_operations pikawdt_fops = {
	.owner		= THIS_MODULE,
	.open		= pikawdt_open,
	.release	= pikawdt_release,
	.write		= pikawdt_write,
	.ioctl		= pikawdt_ioctl,
};

static struct miscdevice pikawdt_miscdev = {
	.minor	= WATCHDOG_MINOR,
	.name	= "watchdog",
	.fops	= &pikawdt_fops,
};

static int __init pikawdt_init(void)
{
	struct device_node *np;
	int ret;

	np = of_find_compatible_node(NULL, NULL, "pika,fpga");
	if (np == NULL) {
		printk(KERN_ERR "pikawdt: Unable to find fpga.\n");
		return -ENOENT;
	}

	pikawdt_fpga = of_iomap(np, 0);

	of_node_put(np);

	if (pikawdt_fpga == NULL) {
		printk(KERN_ERR "pikawdt: Unable to map fpga.\n");
		return -ENOENT;
	}

	ret = misc_register(&pikawdt_miscdev);
	if (ret) {
		iounmap(pikawdt_fpga);
		printk(KERN_ERR "pikawdt: Unable to register miscdev.\n");
		return ret;
	}

	return 0;
}
module_init(pikawdt_init);


static void __exit pikawdt_exit(void)
{
	misc_deregister(&pikawdt_miscdev);

	iounmap(pikawdt_fpga);
}
module_exit(pikawdt_exit);
