/*
 * Based on blktrace code, Copyright (C) 2006 Jens Axboe <axboe@kernel.dk>
 * Moved to utt.c by Tom Zanussi <zanussi@us.ibm.com>, 2006
 * Additional contributions by:
 * Martin Hunt <hunt@redhat.com>, 2007
 * David Wilder <dwilder@us.ibm.com>, 2007
 * Renamed to trace <dwilder.ibm.com>, 2007
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mutex.h>
#include <linux/debugfs.h>
#include <linux/relay_debugfs.h>

static LIST_HEAD(relay_roots);
static LIST_HEAD(relay_dirs);
static DEFINE_MUTEX(relay_mutex);

static int state_open(struct inode *inode, struct file *filp)
{
	filp->private_data = inode->i_private;
	return 0;
}

static ssize_t state_read(struct file *filp, char __user *buffer,
			  size_t count, loff_t *ppos)
{
	struct relay_info *relay = filp->private_data;
	char *buf = "relay not started\n";

	if (relay->state == RELAY_STOPPED)
		buf = "stopped\n";
	else if (relay->state == RELAY_RUNNING)
		buf = "running\n";
	return simple_read_from_buffer(buffer, count, ppos, buf, strlen(buf));
}

static ssize_t state_write(struct file *filp, const char __user *buffer,
			   size_t count, loff_t *ppos)
{
	struct relay_info *relay = filp->private_data;
	char buf[16];
	int ret;

	if (relay->flags & RELAY_DISABLE_STATE)
		return -EINVAL;

	if (count > sizeof(buf) - 1)
		return -EINVAL;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;

	buf[count-1] = '\0';

	if (strcmp(buf, "start") == 0) {
		ret = relay_start(relay);
		if (ret)
			return ret;
	} else if (strcmp(buf, "stop") == 0)
		relay_stop(relay);
	else
		return -EINVAL;

	return count;
}

static struct file_operations state_fops = {
	.owner	= THIS_MODULE,
	.open	= state_open,
	.read	= state_read,
	.write	= state_write,
};

static void remove_root(struct relay_info *relay)
{
	if (relay->root->root && simple_empty(relay->root->root)) {
		debugfs_remove(relay->root->root);
		list_del(&relay->root->list);
		kfree(relay->root);
		relay->root = NULL;
	}
}

static void remove_tree(struct relay_info *relay)
{
	struct list_head *pos, *temp;
	struct relay_dir *dr = NULL;

	mutex_lock(&relay_mutex);
	debugfs_remove(relay->dir);

	list_for_each_safe(pos, temp, &relay_dirs) {
		dr = list_entry(pos, struct relay_dir, relay_dir_list);
		if (dr->ti == relay) {
			list_del(pos);
			kfree(dr);
		}
	}

	if (relay->root) {
		if (--relay->root->users == 0)
			remove_root(relay);
	}

	mutex_unlock(&relay_mutex);
}

/*
 * Creates the relay_root if it's not found.
 */
static struct relay_root *lookup_root(const char *root)
{
	struct list_head *pos;
	struct relay_root *r;

	list_for_each(pos, &relay_roots) {
		r = list_entry(pos, struct relay_root, list);
		if (!strcmp(r->name, root))
			return r;
	}

	r = kzalloc(sizeof(struct relay_root), GFP_KERNEL);
	if (!r)
		return ERR_PTR(-ENOMEM);

	strlcpy(r->name, root, sizeof(r->name));

	r->root = debugfs_create_dir(root, NULL);
	if (IS_ERR(r->root))
		r->root = NULL;
	else
		list_add(&r->list, &relay_roots);

	return r;
}

static struct dentry *create_tree(struct relay_info *relay, const char *root,
				  const char *name)
{
	struct relay_dir *temp;

	if (root == NULL || name == NULL)
		return ERR_PTR(-EINVAL);

	temp = kzalloc(sizeof(struct relay_dir), GFP_KERNEL);
	if ((temp == NULL) || (strlen(name) > RELAY_NAME_SIZE))
		return ERR_PTR(-ENOMEM);

	strlcpy(temp->relay_dir_name, name, sizeof(temp->relay_dir_name));

	mutex_lock(&relay_mutex);

	relay->root = lookup_root(root);
	if (IS_ERR(relay->root)) {
		relay->root = NULL;
		goto err;
	}

	temp->relay_root = relay->root->root;
	temp->relay_dir = debugfs_create_dir(name, relay->root->root);

	if (IS_ERR(temp->relay_dir))
		remove_root(relay);
	else {
		relay->root->users++;
		temp->ti = relay;
		list_add_tail(&temp->relay_dir_list, &relay_dirs);
	}

err:
	mutex_unlock(&relay_mutex);
	return temp->relay_dir;
}

int relay_exists(const char *parent_dir, const char *dir,
					struct relay_info **ti)
{
	struct list_head *pos;
	struct relay_root *r;
	struct relay_dir *temp;

	list_for_each(pos, &relay_roots) {
		r = list_entry(pos, struct relay_root, list);
		if (!strcmp(parent_dir, r->name))
			goto search_dir;
	}
	return RELAY_PARENT_DIR_ABSENT;

 search_dir:
	list_for_each(pos, &relay_dirs) {
		temp = list_entry(pos, struct relay_dir, relay_dir_list);

		if (!strcmp(dir, temp->relay_dir_name)) {
			*ti = temp->ti;
			return RELAY_DIR_EXISTS;
		}
	}
	return RELAY_PARENT_DIR_EXISTS;
}
EXPORT_SYMBOL_GPL(relay_exists);

static int dropped_open(struct inode *inode, struct file *filp)
{
	filp->private_data = inode->i_private;

	return 0;
}

static ssize_t dropped_read(struct file *filp, char __user *buffer,
			    size_t count, loff_t *ppos)
{
	struct relay_info *relay = filp->private_data;
	char buf[16];

	snprintf(buf, sizeof(buf), "%u\n", atomic_read(&relay->dropped));

	return simple_read_from_buffer(buffer, count, ppos, buf, strlen(buf));
}

static struct file_operations dropped_fops = {
	.owner	= THIS_MODULE,
	.open	= dropped_open,
	.read	= dropped_read,
};

static int reset_consumed_open(struct inode *inode, struct file *filp)
{
	filp->private_data = inode->i_private;

	return 0;
}

static ssize_t reset_consumed_write(struct file *filp,
				    const char __user *buffer,
				    size_t count, loff_t *ppos)
{
	int ret = count;
	struct relay_info *relay = filp->private_data;

	mutex_lock(&relay->state_mutex);
	switch (relay->state) {
	case RELAY_RUNNING:
		relay->state = RELAY_STOPPED;
		synchronize_rcu();
		relay_flush(relay->rchan);
		relay_reset_consumed(relay->rchan);
		relay->state = RELAY_RUNNING;
		break;
	case RELAY_STOPPED:
		relay_reset_consumed(relay->rchan);
		break;
	default:
		ret = -EINVAL;
	}
	mutex_unlock(&relay->state_mutex);
	return ret;
}

static struct file_operations reset_consumed_fops = {
	.owner	= THIS_MODULE,
	.open	= reset_consumed_open,
	.write	= reset_consumed_write
};

static int sub_size_open(struct inode *inode, struct file *filp)
{
	filp->private_data = inode->i_private;

	return 0;
}

static ssize_t sub_size_read(struct file *filp, char __user *buffer,
			     size_t count, loff_t *ppos)
{
	struct relay_info *relay = filp->private_data;
	char buf[32];

	snprintf(buf, sizeof(buf), "%zu\n", relay->rchan->subbuf_size);

	return simple_read_from_buffer(buffer, count, ppos, buf, strlen(buf));
}

static struct file_operations sub_size_fops = {
	.owner	= THIS_MODULE,
	.open	= sub_size_open,
	.read	= sub_size_read,
};

static int nr_sub_open(struct inode *inode, struct file *filp)
{
	filp->private_data = inode->i_private;
	return 0;
}

static ssize_t nr_sub_read(struct file *filp, char __user *buffer,
			   size_t count, loff_t *ppos)
{
	struct relay_info *relay = filp->private_data;
	char buf[32];

	snprintf(buf, sizeof(buf), "%zu\n", relay->rchan->n_subbufs);

	return simple_read_from_buffer(buffer, count, ppos, buf, strlen(buf));
}

static struct file_operations nr_sub_fops = {
	.owner	= THIS_MODULE,
	.open	= nr_sub_open,
	.read	= nr_sub_read,
};

/*
 * Keep track of how many times we encountered a full subbuffer, to aid
 * the user space app in telling how many lost events there were.
 */
static int subbuf_start_callback(struct rchan_buf *buf, void *subbuf,
				 void *prev_subbuf, size_t prev_padding)
{
	struct relay_info *relay = buf->chan->private_data;

	if (relay->flags & RELAY_FLIGHT_CHANNEL)
		return 1;

	if (!relay_buf_full(buf))
		return 1;

	atomic_inc(&relay->dropped);

	return 0;
}

static int remove_buf_file_callback(struct dentry *dentry)
{
	debugfs_remove(dentry);

	return 0;
}

static struct dentry *create_buf_file_callback(const char *filename,
					       struct dentry *parent, int mode,
					       struct rchan_buf *buf,
					       int *is_global)
{
	return debugfs_create_file(filename, mode, parent, buf,
				   &relay_file_operations);
}

static struct dentry *create_global_buf_file_callback(const char *filename,
						      struct dentry *parent,
						      int mode,
						      struct rchan_buf *buf,
						      int *is_global)
{
	*is_global = 1;

	return debugfs_create_file(filename, mode, parent, buf,
				   &relay_file_operations);
}

static struct rchan_callbacks relay_callbacks = {
	.subbuf_start		= subbuf_start_callback,
	.create_buf_file	= create_buf_file_callback,
	.remove_buf_file	= remove_buf_file_callback,
};
static struct rchan_callbacks relay_callbacks_global = {
	.subbuf_start		= subbuf_start_callback,
	.create_buf_file	= create_global_buf_file_callback,
	.remove_buf_file	= remove_buf_file_callback,
};

static void remove_controls(struct relay_info *relay)
{
	debugfs_remove(relay->state_file);
	debugfs_remove(relay->dropped_file);
	debugfs_remove(relay->reset_consumed_file);
	debugfs_remove(relay->nr_sub_file);
	debugfs_remove(relay->sub_size_file);
	remove_tree(relay);
}

/*
 * Setup controls for tracing.
 */
static struct relay_info *setup_controls(const char *root,
					 const char *name, u32 flags)
{
	struct relay_info *relay;
	long ret;

	relay = kzalloc(sizeof(*relay), GFP_KERNEL);
	if (!relay) {
		ret = -ENOMEM;
		goto err;
	}

	relay->dir = create_tree(relay, root, name);
	if (IS_ERR(relay->dir)) {
		ret = PTR_ERR(relay->dir);
		relay->dir = NULL;
		goto err;
	}

	relay->state_file = debugfs_create_file("state", 0444, relay->dir,
						 relay, &state_fops);
	if (IS_ERR(relay->state_file)) {
		ret = PTR_ERR(relay->state_file);
		relay->state_file = NULL;
		goto err;
	}

	if (!(flags & RELAY_FLIGHT_CHANNEL)) {
		relay->dropped_file = debugfs_create_file("dropped", 0444,
							  relay->dir, relay,
							  &dropped_fops);
		if (IS_ERR(relay->dropped_file)) {
			ret = PTR_ERR(relay->dropped_file);
			relay->dropped_file = NULL;
			goto err;
		}
	}

	if (flags & RELAY_FLIGHT_CHANNEL) {
		relay->reset_consumed_file = debugfs_create_file("rewind", 0444,
							relay->dir, relay,
							&reset_consumed_fops);
		if (IS_ERR(relay->reset_consumed_file)) {
			ret = PTR_ERR(relay->reset_consumed_file);
			relay->reset_consumed_file = NULL;
			goto err;
		}
	}

	relay->nr_sub_file = debugfs_create_file("nr_sub", 0444,
						 relay->dir, relay,
						 &nr_sub_fops);
	if (IS_ERR(relay->nr_sub_file)) {
		ret = PTR_ERR(relay->nr_sub_file);
		relay->nr_sub_file = NULL;
		goto err;
	}

	relay->sub_size_file = debugfs_create_file("sub_size", 0444,
						   relay->dir, relay,
						   &sub_size_fops);
	if (IS_ERR(relay->sub_size_file)) {
		ret = PTR_ERR(relay->sub_size_file);
		relay->sub_size_file = NULL;
		goto err;
	}

	return relay;
err:
	if (relay) {
		remove_controls(relay);
		kfree(relay);
	}

	return ERR_PTR(ret);
}

static int relay_setup_channel(struct relay_info *relay, u32 buf_size,
				u32 buf_nr, u32 flags)
{
	if (!buf_size || !buf_nr)
		return -EINVAL;

	if (flags & RELAY_GLOBAL_CHANNEL)
		relay->rchan = relay_open("trace", relay->dir, buf_size,
					  buf_nr, &relay_callbacks_global,
					  relay);
	else
		relay->rchan = relay_open("trace", relay->dir, buf_size,
					  buf_nr, &relay_callbacks, relay);

	if (!relay->rchan)
		return -ENOMEM;

	relay->flags = flags;
	relay->state = RELAY_SETUP;

	return 0;
}

/**
 *	relay_setup - create a new relay relay handle
 *	@root: The root directory name to place relay directories.
 *	@name: Relay debugfs directory name, created in @root
 *	@buf_size: size of the relay sub-buffers
 *	@buf_nr: number of relay sub-buffers
 *	@flags: Option selection (see relay channel flags definitions)
 *
 *	returns a relay_info handle or NULL, if setup failed.
 *
 *	The @root is created (if needed) in the root of the debugfs.
 *	The default values when flags=0 are: use per-CPU buffering,
 *	use non-overwrite mode. See Documentation/filesystems/relay.txt for
 *	details.
 */
struct relay_info *relay_setup(const char *root, const char *name,
			       u32 buf_size, u32 buf_nr, u32 flags)
{
	struct relay_info *relay;

	relay = setup_controls(root, name, flags);
	if (IS_ERR(relay))
		return relay;

	relay->buf_size = buf_size;
	relay->buf_nr = buf_nr;
	relay->flags = flags;
	mutex_init(&relay->state_mutex);
	relay->state = RELAY_SETUP;

	return relay;
}
EXPORT_SYMBOL_GPL(relay_setup);

/**
 *	relay_start - start tracing
 *	@relay: relay handle to start.
 *
 *	returns 0 if successful.
 */
int relay_start(struct relay_info *relay)
{
	/*
	 * For starting a relay, we can transition from a setup or stopped
	 * relay.
	 */
	if (relay->state == RELAY_RUNNING)
		return -EINVAL;

	mutex_lock(&relay->state_mutex);
	if (relay->state == RELAY_SETUP) {
		int ret;

		ret = relay_setup_channel(relay, relay->buf_size,
					  relay->buf_nr, relay->flags);
		if (ret) {
			mutex_unlock(&relay->state_mutex);
			return ret;
		}
	}

	relay->state = RELAY_RUNNING;
	mutex_unlock(&relay->state_mutex);

	return 0;
}
EXPORT_SYMBOL_GPL(relay_start);

/**
 *	relay_stop - stop tracing
 *	@relay: relay handle to stop.
 *
 */
int relay_stop(struct relay_info *relay)
{
	int ret = -EINVAL;

	/*
	 * For stopping a relay, the state must be running
	 */
	mutex_lock(&relay->state_mutex);
	if (relay->state == RELAY_RUNNING) {
		relay->state = RELAY_STOPPED;
		/*
		 * wait for all cpus to see the change in
		 * state before continuing
		 */
		synchronize_sched();
		relay_flush(relay->rchan);
		ret = 0;
	}
	mutex_unlock(&relay->state_mutex);
	return ret;
}
EXPORT_SYMBOL_GPL(relay_stop);

static void relay_cleanup_channel(struct relay_info *relay)
{
	relay_stop(relay);
	relay_close(relay->rchan);
	relay->rchan = NULL;
}

/**
 *	relay_cleanup - destroys the relay channel, control files and dir
 *	@relay: relay handle to cleanup
 */
void relay_cleanup(struct relay_info *relay)
{
	relay_cleanup_channel(relay);
	remove_controls(relay);
	kfree(relay);
}
EXPORT_SYMBOL_GPL(relay_cleanup);

/**
 *	relay_cleanup_all - Removes all relay debugfs directories in parent_dir
 *	@parent_dir: Name of the parent directory
 */
void relay_cleanup_all(const char *parent_dir)
{
	struct list_head *pos, *pos_temp;
	struct relay_dir *temp;

	list_for_each_safe(pos, pos_temp, &relay_dirs) {
		temp = list_entry(pos, struct relay_dir, relay_dir_list);
		if (!strncmp(parent_dir, temp->relay_root->d_iname, \
							strlen(parent_dir)))
			relay_cleanup(temp->ti);
	}
}
EXPORT_SYMBOL_GPL(relay_cleanup_all);

/*
 * Send formatted data to relay debugfs channel.
 */
static int relay_printf(struct relay_info *trace, const char *format,
								va_list ap)
{
	va_list aq;
	char *record;
	int len, ret = 0;

	if (relay_running(trace)) {
		va_copy(aq, ap);
		len = vsnprintf(NULL, 0, format, aq);
		va_end(aq);
		record = relay_reserve(trace->rchan, ++len);
		if (record)
			ret = vsnprintf(record, len, format, ap);
	}
	return ret;
}

static inline int init_relay_interface(struct relay_printk_data *tpk)
{
	int ret = 0;
	tpk->exists = relay_exists(tpk->parent_dir, tpk->dir, &tpk->ti);

	switch (tpk->exists) {

	case RELAY_PARENT_DIR_EXISTS:
	case RELAY_PARENT_DIR_ABSENT:
		if (!tpk->buf_size)
			tpk->buf_size = DEFAULT_RELAY_BUF_SIZE;
		if (!tpk->sub_buf_size)
			tpk->sub_buf_size = DEFAULT_RELAY_SUB_BUF_NR;
		tpk->ti = relay_setup(tpk->parent_dir, tpk->dir,
				tpk->buf_size, tpk->sub_buf_size, tpk->flags);
		printk(KERN_INFO "Trace interface %s setup\n",
							tpk->ti->dir->d_iname);
		if (IS_ERR(tpk->ti)) {
			printk(KERN_ERR "Error initialising %s interface\n",
							tpk->ti->dir->d_iname);
			return -EPERM;
		}
	/* Fall through */
	case RELAY_DIR_EXISTS:
		if (tpk->ti->state == RELAY_SETUP)
			ret = relay_start(tpk->ti);
		else
			ret = -EPERM;
	}

	return 0;
}

/**
 * relay_printk - Output a string to debugfs mount 'directly' using 'trace'
 * @tpk: Structure containing info such as parent_dir and directory
 * @format: String containing format string specifiers
 * @ap: List of arguments
 */
int relay_printk(struct relay_printk_data *tpk, char *format, ...)
{
	int ret = 0;
	va_list(ap);
	unsigned long flags = 0;

	va_start(ap, format);

	ret = init_relay_interface(tpk);
	if (unlikely(ret))
		return ret;

	/* Now do the actual printing */
	/* Take an RCU Lock over the relay_info state */
	rcu_read_lock();
	/* Take a spinlock for the global buffer used by relay */
	if (tpk->flags & RELAY_GLOBAL_CHANNEL)
		spin_lock_irqsave(&tpk->ti->relay_lock, flags);
	ret = relay_printf(tpk->ti, format, ap);
	if (tpk->flags & RELAY_GLOBAL_CHANNEL)
		spin_unlock_irqrestore(&tpk->ti->relay_lock, flags);
	rcu_read_unlock();

	va_end(ap);
	return ret;
}
EXPORT_SYMBOL(relay_printk);

/**
 * relay_dump - Output binary into debugfs mount 'directly' using 'trace'
 * @tpk: Structure containing info such as parent_dir and directory
 * @output: Data that needs to be output
 * @output_len: Length of the output data
 */
int relay_dump(struct relay_printk_data *tpk, const void *output,
							const int output_len)
{
	char *record;
	unsigned long flags = 0;
	int ret = 0;

	ret = init_relay_interface(tpk);
	if (unlikely(ret))
		return ret;

	/* Now do the actual printing */
	rcu_read_lock();
	/* Take a spinlock for the global buffer used by relay */
	if (tpk->flags & RELAY_GLOBAL_CHANNEL)
		spin_lock_irqsave(&tpk->ti->relay_lock, flags);
	record = relay_reserve(tpk->ti->rchan, output_len);

	if (record && relay_running(tpk->ti))
		memcpy(record, output, output_len);
	else {
		if (record)
			ret = -EPERM;
		else
			ret = -ENOMEM;
	}
	if (tpk->flags & RELAY_GLOBAL_CHANNEL)
		spin_unlock_irqrestore(&tpk->ti->relay_lock, flags);
	rcu_read_unlock();

	return ret;
}
EXPORT_SYMBOL(relay_dump);
