/*
 * An example of using 'relay debugfs' in a kprobes module
 *
 * Copyright (C) 2007 IBM Inc.
 *
 * David Wilder <dwilder@us.ibm.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * -------
 * This module creates a 'relay debugfs' channel and places a kprobe
 * on the function do_fork(). The value of current->pid is written to
 * the 'relay debugfs' channel each time the kprobe is hit..
 *
 * How to run the example:
 * $ mount -t debugfs /debug
 * $ insmod fork_trace.ko
 *
 * To view the data produced by the module:
 * $ cat /debug/relay_debugfs_example/do_fork/trace0
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/relay_debugfs.h>

#define USE_GLOBAL_BUFFERS 1
#define USE_FLIGHT 1

#define PROBE_POINT "do_fork"

static struct kprobe kp;
static struct relay_info *kprobes_relay;

#ifdef USE_GLOBAL_BUFFERS
static DEFINE_SPINLOCK(relay_debugfs_lock);
#endif

/*
 * Send formatted trace data to 'relay debugfs' channel.
 * @note Preemption must be disabled to use this.
 */
static void relay_debugfs_printf(struct relay_info *relay, const char *format, ...)
{
	va_list ap, aq;
	char *record;
	unsigned long flags;
	int len;

	if (!relay)
		return;

#ifdef USE_GLOBAL_BUFFERS
	spin_lock_irqsave(&relay_debugfs_lock, flags);
#endif
	if (relay_running(relay)) {
		va_start(ap, format);
		va_copy(aq, ap);
		len = vsnprintf(NULL, 0, format, aq);
		va_end(aq);
		record = relay_reserve(relay->rchan, ++len);
		if (record)
			vsnprintf(record, len, format, ap);
		va_end(ap);
	}
#ifdef USE_GLOBAL_BUFFERS
	spin_unlock_irqrestore(&relay_debugfs_lock, flags);
#endif
}

static int handler_pre(struct kprobe *p, struct pt_regs *regs)
{
	rcu_read_lock();
	relay_debugfs_printf(kprobes_relay, "%d\n", current->pid);
	rcu_read_unlock();
	return 0;
}

int init_module(void)
{
	int ret;
	u32 flags = 0;

#ifdef USE_GLOBAL_BUFFERS
	flags |= RELAY_GLOBAL_CHANNEL;
#endif

#ifdef USE_FLIGHT
	flags |= RELAY_FLIGHT_CHANNEL;
#endif

	/* setup the relay */
	kprobes_relay = relay_setup("relay_example", PROBE_POINT,
				     1024, 8, flags);
	if (IS_ERR(kprobes_relay))
		return PTR_ERR(kprobes_relay);

	relay_start(kprobes_relay);

	/* setup the kprobe */
	kp.pre_handler = handler_pre;
	kp.post_handler = NULL;
	kp.fault_handler = NULL;
	kp.symbol_name = PROBE_POINT;
	ret = register_kprobe(&kp);
	if (ret) {
		printk(KERN_ERR "fork_trace: register_kprobe failed\n");
		return ret;
	}
	return 0;
}

void cleanup_module(void)
{
	unregister_kprobe(&kp);
	relay_stop(kprobes_relay);
	relay_cleanup(kprobes_relay);
}
MODULE_LICENSE("GPL");
